"""This module defines customized metric aggregation functions."""

import itertools
import math
from multiprocessing import Pool

import numpy as np

from fibber.metrics.editing_distance import EditingDistance

DEFAULT_USE_SIM_THRESHOLD = 0.85
DEFAULT_GPT2_PPL_THRESHOLD = 5


def paraphrase_classification_accuracy_agg_fn(use_sim, ppl_score):
    """This function makes a aggregation function for BERT classification metric.

    The aggregation function outputs the after attack accuracy of the BERT classifier.

    Args:
        use_sim (float): the threshold for ``USESemanticSimilarity`` metric. The adversarial
            example should have the USE similarity higher than this threshold.
        ppl_score (float): the threshold for ``GPT2GrammarQuality`` metric. The adversarial example
            should have the perplexity ratio measured by GPT2 lower than this threshold.
    """

    def agg_fn(data_record):
        if data_record["original_text_metrics"]["BertClfPrediction"] != data_record["label"]:
            return 0
        for item in data_record["paraphrase_metrics"]:
            if (item["BertClfPrediction"] != data_record["label"]
                    and item["GPT2GrammarQuality"] < ppl_score
                    and item["USESemanticSimilarity"] > use_sim):
                return 0
        return 1

    return agg_fn


def editing_distance_element_fn(x):
    editing_distance_metric = EditingDistance()
    return editing_distance_metric.measure_example(x[0], x[1])


pool = Pool(8)


def pairwise_editing_distance_fn(data_record):
    """Compute the average pairwise editing distance metric."""
    global pool
    paraphrases = None
    for k, v in data_record.items():
        if k in ["text0_paraphrases", "text1_paraphrases"]:
            paraphrases = v
            break

    assert paraphrases is not None

    tuples = list(itertools.combinations(paraphrases, 2))

    distance = pool.map(editing_distance_element_fn, tuples)
    return float(np.mean(distance))


def get_best_adv_by_sim(data_record):
    """Find the adversarial example with best similarity.

    Args:
        data_record (dict): a data record with paraphrases and metrics.
    Returns:
         (dict): the metrics of the best adversarial example. None if no legitimate adversarial
            example is found.
    """
    if data_record["original_text_metrics"]["BertClfPrediction"] != data_record["label"]:
        return None
    best_score = -1
    best_metrics = None
    for metrics in data_record["paraphrase_metrics"]:
        if (metrics["BertClfPrediction"] == data_record["label"]
                or metrics["GPT2GrammarQuality"] > DEFAULT_GPT2_PPL_THRESHOLD
                or metrics["USESemanticSimilarity"] < DEFAULT_USE_SIM_THRESHOLD):
            continue
        if metrics["USESemanticSimilarity"] > best_score:
            best_score = metrics["USESemanticSimilarity"]
            best_metrics = metrics
    return best_metrics


def get_best_adv_by_ppl(data_record):
    """Find the adversarial example with lowest perplexity.

    Args:
        data_record (dict): a data record with paraphrases and metrics.
    Returns:
         (dict): the metrics of the best adversarial example. None if no legitimate adversarial
            example is found.
    """
    if data_record["original_text_metrics"]["BertClfPrediction"] != data_record["label"]:
        return None
    best_score = 1e8
    best_metrics = None
    for metrics in data_record["paraphrase_metrics"]:
        if (metrics["BertClfPrediction"] == data_record["label"]
                or metrics["GPT2GrammarQuality"] > DEFAULT_GPT2_PPL_THRESHOLD
                or metrics["USESemanticSimilarity"] < DEFAULT_USE_SIM_THRESHOLD):
            continue
        if metrics["GPT2GrammarQuality"] < best_score:
            best_score = metrics["GPT2GrammarQuality"]
            best_metrics = metrics
    return best_metrics


def get_best_adv_metric_fn(get_best_adv_fn, metric_name):
    """Returns an aggregation function that extracts the value of a specified metric for the best
    adversarial example.

    The aggregation function returns NaN if no legitimate adversarial example is found.

    Args:
        get_best_adv_fn (fn): a function that returns the metric dict of the best adversarial
            example.
        metric_name (str): a metric name.
    Returns:
        (fn): an aggregation function that takes data_record as an input.
    """
    def agg_fn(data_record):
        best_metrics = get_best_adv_fn(data_record)
        if best_metrics is not None:
            return best_metrics[metric_name]
        return math.nan

    return agg_fn


customized_metric_aggregation_fn_dict = {
    "3_ParaphraseAcc_use0.90_ppl2":
        paraphrase_classification_accuracy_agg_fn(use_sim=0.90, ppl_score=2),
    "4_ParaphraseAcc_use0.85_ppl5":
        paraphrase_classification_accuracy_agg_fn(use_sim=0.85, ppl_score=5),
    "pairwise_edit_distance": pairwise_editing_distance_fn,
    "best_sim_adv_GPT2GrammarQuality":
        get_best_adv_metric_fn(get_best_adv_by_sim, "GPT2GrammarQuality"),
    "best_sim_adv_USESemanticSimilarity":
        get_best_adv_metric_fn(get_best_adv_by_sim, "USESemanticSimilarity"),
    "best_sim_adv_EditingDistance":
        get_best_adv_metric_fn(get_best_adv_by_sim, "EditingDistance"),
    "best_sim_adv_GloveSemanticSimilarity":
        get_best_adv_metric_fn(get_best_adv_by_sim, "GloVeSemanticSimilarity"),
    "best_ppl_adv_GPT2GrammarQuality":
        get_best_adv_metric_fn(get_best_adv_by_ppl, "GPT2GrammarQuality"),
    "best_ppl_adv_USESemanticSimilarity":
        get_best_adv_metric_fn(get_best_adv_by_ppl, "USESemanticSimilarity"),
    "best_ppl_adv_EditingDistance":
        get_best_adv_metric_fn(get_best_adv_by_ppl, "EditingDistance"),
    "best_ppl_adv_GloveSemanticSimilarity":
        get_best_adv_metric_fn(get_best_adv_by_ppl, "GloVeSemanticSimilarity"),
}

customized_metric_for_nun_wins = [
    ("3_ParaphraseAcc_use0.90_ppl2", "L"),
    ("4_ParaphraseAcc_use0.85_ppl5", "L"),
    ("best_sim_adv_USESemanticSimilarity", "H"),
    ("best_sim_adv_GPT2GrammarQuality", "L"),
    ("best_ppl_adv_USESemanticSimilarity", "H"),
    ("best_ppl_adv_GPT2GrammarQuality", "L"),
]



"""This metric outputs the BERT classifier prediction of the paraphrased text."""

import os

import numpy as np
import torch
import torch.nn.functional as F
import tqdm
import transformers
from torch.utils.tensorboard import SummaryWriter
from transformers import BertForSequenceClassification, BertTokenizerFast

from fibber import get_root_dir, log, resources
from fibber.datasets import DatasetForBert
from fibber.metrics.metric_base import MetricBase

logger = log.setup_custom_logger(__name__)


def get_optimizer(optimizer_name, lr, decay, train_step, params, warmup=1000):
    """Create an optimizer and schedule of learning rate for parameters.

    Args:
        optimizer_name (str): choose from ``["adam", "sgd", "adamw"]``.
        lr (float): learning rate.
        decay (float): weight decay.
        train_step (int): number of training steps.
        params (list): a list of parameters in the model.
        warmup (int): number of warm up steps.

    Returns:
        A torch optimizer and a scheduler.
    """
    if optimizer_name == "adam":
        opt = torch.optim.Adam(params=params, lr=lr, weight_decay=decay)
    elif optimizer_name == "sgd":
        opt = torch.optim.SGD(params=params, lr=lr, weight_decay=decay)
    elif optimizer_name == "adamw":
        opt = torch.optim.AdamW(params=params, lr=lr, weight_decay=decay)
    else:
        assert 0, "unkown optimizer"

    sche = transformers.get_linear_schedule_with_warmup(
        opt, num_warmup_steps=warmup, num_training_steps=train_step)

    return opt, sche


def run_evaluate(model, dataloader_iter, eval_steps, summary, global_step, device):
    """Evaluate a model and add error rate and validation loss to Tensorboard.

    Args:
        model (transformers.BertForSequenceClassification): a BERT classification model.
        dataloader_iter (torch.IterableDataset): an iterator of a torch.IterableDataset.
        eval_steps (int): number of training steps.
        summary (torch.utils.tensorboard.SummaryWriter): a Tensorboard SummaryWriter object.
        global_step (int): current training steps.
        device (torch.Device): the device where the model in running on.
    """
    model.eval()

    correct, count = 0, 0
    loss_list = []

    for v_step in range(eval_steps):
        seq, mask, tok_type, label = next(dataloader_iter)
        seq = seq.to(device)
        mask = mask.to(device)
        tok_type = tok_type.to(device)
        label = label.to(device)

        with torch.no_grad():
            outputs = model(seq, mask, tok_type, labels=label)
            loss, logits = outputs[:2]
            loss_list.append(loss.detach().cpu().numpy())
            count += seq.size(0)
            correct += (logits.argmax(dim=1).eq(label)
                        .float().sum().detach().cpu().numpy())

    summary.add_scalar("clf_val/error_rate", 1 - correct / count, global_step)
    summary.add_scalar("clf_val/loss", np.mean(loss_list), global_step)

    model.train()


def load_or_train_bert_clf(model_init,
                           dataset_name,
                           trainset,
                           testset,
                           bert_clf_steps,
                           bert_clf_bs,
                           bert_clf_lr,
                           bert_clf_optimizer,
                           bert_clf_weight_decay,
                           bert_clf_period_summary,
                           bert_clf_period_val,
                           bert_clf_period_save,
                           bert_clf_val_steps,
                           device):
    """Train BERT classification model on a dataset.

    The trained model will be stored at ``<fibber_root_dir>/bert_clf/<dataset_name>/``. If there's
    a saved model, load and return the model. Otherwise, train the model using the given data.

    Args:
        model_init (str): pretrained model name. Choose from ``["bert-base-cased",
            "bert-base-uncased", "bert-large-cased", "bert-large-uncased"]``.
        dataset_name (str): the name of the dataset. This is also the dir to save trained model.
        trainset (dict): a fibber dataset.
        testset (dict): a fibber dataset.
        bert_clf_steps (int): steps to train a classifier.
        bert_clf_bs (int): the batch size.
        bert_clf_lr (float): the learning rate.
        bert_clf_optimizer (str): the optimizer name.
        bert_clf_weight_decay (float): the weight decay.
        bert_clf_period_summary (int): the period in steps to write training summary.
        bert_clf_period_val (int): the period in steps to run validation and write validation
            summary.
        bert_clf_period_save (int): the period in steps to save current model.
        bert_clf_val_steps (int): number of batched in each validation.
        device (torch.Device): the device to run the model.

    Returns:
        (transformers.BertForSequenceClassification): a torch BERT model.
    """
    model_dir = os.path.join(get_root_dir(), "bert_clf", dataset_name, )
    ckpt_path = os.path.join(model_dir, model_init + "-%04dk.pt" %
                             (bert_clf_steps // 1000))

    if os.path.exists(ckpt_path):
        logger.info("Load BERT classifier from %s.", ckpt_path)
        model = torch.load(ckpt_path)
        model.eval()
        model.to(device)
        return model

    num_labels = len(trainset["label_mapping"])

    model = BertForSequenceClassification.from_pretrained(
        resources.get_transformers(model_init), num_labels=num_labels).to(device)
    model.train()

    logger.info("Use %s tokenizer and classifier.", model_init)
    logger.info("Num labels: %s", num_labels)

    summary = SummaryWriter(os.path.join(model_dir, "summary"))

    dataloader = torch.utils.data.DataLoader(
        DatasetForBert(trainset, model_init, bert_clf_bs), batch_size=None, num_workers=2)

    dataloader_val = torch.utils.data.DataLoader(
        DatasetForBert(testset, model_init, bert_clf_bs), batch_size=None, num_workers=1)
    dataloader_val_iter = iter(dataloader_val)

    params = model.parameters()

    opt, sche = get_optimizer(
        bert_clf_optimizer, bert_clf_lr, bert_clf_weight_decay, bert_clf_steps, params)

    global_step = 0
    correct_train, count_train = 0, 0
    for seq, mask, tok_type, label in tqdm.tqdm(dataloader, total=bert_clf_steps):
        global_step += 1
        seq = seq.to(device)
        mask = mask.to(device)
        tok_type = tok_type.to(device)
        label = label.to(device)

        outputs = model(seq, mask, tok_type, labels=label)
        loss, logits = outputs[:2]

        count_train += seq.size(0)
        correct_train += (logits.argmax(dim=1).eq(label)
                          .float().sum().detach().cpu().numpy())

        opt.zero_grad()
        loss.backward()
        opt.step()
        sche.step()

        if global_step % bert_clf_period_summary == 0:
            summary.add_scalar("clf_train/loss", loss, global_step)
            summary.add_scalar("clf_train/error_rate", 1 - correct_train / count_train,
                               global_step)
            correct_train, count_train = 0, 0

        if global_step % bert_clf_period_val == 0:
            run_evaluate(model, dataloader_val_iter,
                         bert_clf_val_steps, summary, global_step, device)

        if global_step % bert_clf_period_save == 0 or global_step == bert_clf_steps:
            model.eval()
            model.to(torch.device("cpu"))
            ckpt = os.path.join(model_dir, model_init + "-%04dk.pt" % (global_step // 1000))
            torch.save(model, ckpt)
            logger.info("BERT classifier saved at %s.", ckpt)
            model.to(device)
            model.train()

        if global_step >= bert_clf_steps:
            break
    model.eval()
    return model


class BertClfPrediction(MetricBase):
    """BERT classifier prediction on paraphrases.

    This metric is special, it does not compare the original and paraphrased sentence.
    Instead, it outputs the classifier prediction on paraphrases. So we should not compute
    mean or std on this metric.

    Args:
        dataset_name (str): the name of the dataset.
        trainset (dict): a fibber dataset.
        testset (dict): a fibber dataset.
        bert_gpu_id (int): the gpu id for BERT model. Set -1 to use CPU.
        bert_clf_steps (int): steps to train a classifier.
        bert_clf_bs (int): the batch size.
        bert_clf_lr (float): the learning rate.
        bert_clf_optimizer (str): the optimizer name.
        bert_clf_weight_decay (float): the weight decay in the optimizer.
        bert_clf_period_summary (int): the period in steps to write training summary.
        bert_clf_period_val (int): the period in steps to run validation and write validation
            summary.
        bert_clf_period_save (int): the period in steps to save current model.
        bert_clf_val_steps (int): number of batched in each validation.
    """

    def __init__(self, dataset_name, trainset, testset, bert_gpu_id=-1,
                 bert_clf_steps=20000, bert_clf_bs=32, bert_clf_lr=0.00002,
                 bert_clf_optimizer="adamw", bert_clf_weight_decay=0.001,
                 bert_clf_period_summary=100, bert_clf_period_val=500,
                 bert_clf_period_save=5000, bert_clf_val_steps=10, **kargs):

        super(BertClfPrediction, self).__init__()

        if trainset["cased"]:
            model_init = "bert-base-cased"
            logger.info(
                "Use cased model in BERT classifier prediction metric.")
        else:
            model_init = "bert-base-uncased"
            logger.info(
                "Use uncased model in BERT classifier prediction metric.")

        self._tokenizer = BertTokenizerFast.from_pretrained(
            resources.get_transformers(model_init), do_lower_case="uncased" in model_init)

        if bert_gpu_id == -1:
            logger.warning("BERT metric is running on CPU.")
            self._device = torch.device("cpu")
        else:
            logger.info("BERT metric is running on GPU %d.", bert_gpu_id)
            self._device = torch.device("cuda:%d" % bert_gpu_id)

        self._model = load_or_train_bert_clf(
            model_init=model_init,
            dataset_name=dataset_name,
            trainset=trainset,
            testset=testset,
            bert_clf_steps=bert_clf_steps,
            bert_clf_lr=bert_clf_lr,
            bert_clf_bs=bert_clf_bs,
            bert_clf_optimizer=bert_clf_optimizer,
            bert_clf_weight_decay=bert_clf_weight_decay,
            bert_clf_period_summary=bert_clf_period_summary,
            bert_clf_period_val=bert_clf_period_val,
            bert_clf_period_save=bert_clf_period_save,
            bert_clf_val_steps=bert_clf_val_steps,
            device=self._device)

    def predict_dist_batch(self, text, context):
        """Get log probability of the BERT prediction.

        Args:
            text ([str]): a list of strs to be classified.
            context (str or None):
                On regular classification datasets, context is None.
                On NLI datasets, context contains the premise.

        Returns:
            (np.array): a numpy array of the unnormalized log probability over on each category.
        """
        if context is None:
            batch_input = self._tokenizer(text=text, padding=True, max_length=200,
                                          truncation=True)
        else:
            batch_input = self._tokenizer(
                text=[context] * len(text), text_pair=text, padding=True, max_length=200,
                truncation=True)

        with torch.no_grad():
            res = F.log_softmax(self._model(
                input_ids=torch.tensor(batch_input["input_ids"]).to(self._device),
                token_type_ids=torch.tensor(batch_input["token_type_ids"]).to(self._device),
                attention_mask=torch.tensor(batch_input["attention_mask"]).to(self._device)
            )[0], dim=1).detach().cpu().numpy()
        return res

    def predict_batch(self, text, context):
        """Get the prediction of the BERT prediction.

        Args:
            text ([str]): a list of strs to be classified.
            context (str or None):
                On regular classification datasets, context is None.
                On NLI datasets, context contains the premise.

        Returns:
            (np.array): the prediction of the classifier.
        """

        return np.argmax(self.predict_dist_batch(text=text, context=context), axis=1)

    def measure_batch(self, origin, paraphrases, data_record=None, paraphrase_field="text0"):
        """Measure the metric on a batch of paraphrases.

        Args:
            origin (str): the original text.
            paraphrases (list): a set of paraphrases.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (list): a list containing the metric for each paraphrase.
        """
        res = self.predict_batch(paraphrases,
                                 data_record["text0"] if paraphrase_field == "text1" else None)
        return [int(x) for x in res]

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        if paraphrase_field == "text0":
            return int(self.predict_batch([paraphrase], None)[0])
        else:
            assert paraphrase_field == "text1"
            return int(self.predict_batch([paraphrase], data_record["text0"])[0])



import argparse
import datetime
import os

from fibber import log
from fibber.benchmark.benchmark_utils import update_detailed_result
from fibber.benchmark.customized_metric_aggregation import customized_metric_aggregation_fn_dict
from fibber.datasets import get_dataset, subsample_dataset
from fibber.metrics import MetricBundle, aggregate_metrics, compute_metrics
from fibber.paraphrase_strategies import (
    BertSamplingStrategy, IdentityStrategy, RandomStrategy, TextFoolerStrategy)

logger = log.setup_custom_logger(__name__)
log.remove_logger_tf_handler(logger)

parser = argparse.ArgumentParser()

parser.add_argument("--dataset", type=str, default="ag")
parser.add_argument("--output_dir", type=str, default=None)
parser.add_argument("--num_paraphrases_per_text", type=int, default=20)
parser.add_argument("--subsample_testset", type=int, default=1000)
parser.add_argument("--strategy", type=str, default="RandomStrategy")
parser.add_argument("--strategy_gpu_id", type=int, default=-1)

# metric args
parser.add_argument("--gpt2_gpu_id", type=int, default=-1)
parser.add_argument("--bert_gpu_id", type=int, default=-1)
parser.add_argument("--use_gpu_id", type=int, default=-1)
parser.add_argument("--bert_clf_steps", type=int, default=20000)

RandomStrategy.add_parser_args(parser)
IdentityStrategy.add_parser_args(parser)
TextFoolerStrategy.add_parser_args(parser)
BertSamplingStrategy.add_parser_args(parser)

G_EXP_NAME = None


def get_output_filename(arg_dict, prefix="", suffix=""):
    """Returns a string like ``<prefix>-<experiment_name>-<suffix>``.

    This function is used to construct file names for an experiment. This function ensures that
    all file names for the same experiment has the same experiment name.

    The experiment name is ``<dataset_name>-<paraphrase_strategy_name>-<date>-<time>``.
    """
    global G_EXP_NAME
    if G_EXP_NAME is None:
        G_EXP_NAME = (arg_dict["dataset"] + "-" + arg_dict["strategy"] + "-"
                      + datetime.datetime.now().strftime("%m%d-%H%M%S"))
    return prefix + G_EXP_NAME + suffix


def get_strategy(arg_dict, dataset_name, strategy_name, strategy_gpu_id,
                 output_dir, metric_bundle):
    """Take the strategy name and construct a strategy object."""
    if strategy_name == "RandomStrategy":
        return RandomStrategy(arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle)
    if strategy_name == "IdentityStrategy":
        return IdentityStrategy(
            arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle)
    if strategy_name == "TextFoolerStrategy":
        return TextFoolerStrategy(
            arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle)
    if strategy_name == "BertSamplingStrategy":
        return BertSamplingStrategy(
            arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle)
    else:
        assert 0


def benchmark(arg_dict, dataset_name, trainset, testset, paraphrase_set):
    """Run benchmark on the given dataset.

    Args:
        arg_dict (dict): all args from the commandline.
        dataset_name (str): the name of the dataset.
        trainset (dict): a fibber dataset.
        testset (dict): a fibber dataset.
        paraphrase_set (dict): a fibber dataset.
    """
    logger.info("Build metric bundle.")

    metric_bundle = MetricBundle(
        use_bert_clf_prediction=True,
        use_gpu_id=arg_dict["use_gpu_id"], gpt2_gpu_id=arg_dict["gpt2_gpu_id"],
        bert_gpu_id=arg_dict["bert_gpu_id"], dataset_name=dataset_name,
        trainset=trainset, testset=testset,
        bert_clf_steps=arg_dict["bert_clf_steps"])

    paraphrase_strategy = get_strategy(arg_dict, arg_dict["dataset"], arg_dict["strategy"],
                                       arg_dict["strategy_gpu_id"], arg_dict["output_dir"],
                                       metric_bundle)
    paraphrase_strategy.fit(trainset)

    tmp_output_filename = os.path.join(
        arg_dict["output_dir"], get_output_filename(arg_dict, suffix="-tmp.json"))
    logger.info("Write paraphrase temporary results in %s.", tmp_output_filename)
    results = paraphrase_strategy.paraphrase_dataset(
        paraphrase_set, arg_dict["num_paraphrases_per_text"], tmp_output_filename)

    output_filename = os.path.join(
        arg_dict["output_dir"], get_output_filename(arg_dict, suffix="-with-metric.json"))

    results = compute_metrics(metric_bundle, results=results, output_filename=output_filename)

    aggregated_result = aggregate_metrics(
        dataset_name, str(paraphrase_strategy), G_EXP_NAME, results,
        customized_metric_aggregation_fn_dict)
    update_detailed_result(aggregated_result)


def main():
    arg_dict = vars(parser.parse_args())
    assert arg_dict["output_dir"] is not None
    os.makedirs(arg_dict["output_dir"], exist_ok=True)
    os.makedirs(os.path.join(arg_dict["output_dir"], "log"), exist_ok=True)

    log.add_file_handler(
        logger, os.path.join(arg_dict["output_dir"], "log",
                             get_output_filename(arg_dict, suffix=".log")))
    log.remove_logger_tf_handler(logger)

    trainset, testset = get_dataset(arg_dict["dataset"])
    paraphrase_set = subsample_dataset(testset, arg_dict["subsample_testset"])
    logger.info("Subsample test set to %d.", arg_dict["subsample_testset"])
    benchmark(arg_dict, arg_dict["dataset"], trainset, testset, paraphrase_set)


if __name__ == "__main__":
    main()



from fibber.metrics.bert_clf_prediction import BertClfPrediction
from fibber.metrics.editing_distance import EditingDistance
from fibber.metrics.glove_semantic_similarity import GloVeSemanticSimilarity
from fibber.metrics.gpt2_grammar_quality import GPT2GrammarQuality
from fibber.metrics.metric_base import MetricBase
from fibber.metrics.metric_utils import MetricBundle, aggregate_metrics, compute_metrics
from fibber.metrics.use_semantic_similarity import USESemanticSimilarity

__all__ = [
    "BertClfPrediction",
    "EditingDistance",
    "GloVeSemanticSimilarity",
    "GPT2GrammarQuality",
    "MetricBase",
    "aggregate_metrics",
    "compute_metrics",
    "USESemanticSimilarity",
    "MetricBundle"]



from fibber.resources.resource_utils import (
    get_corenlp, get_glove_emb, get_nltk_data, get_stopwords, get_transformers,
    get_universal_sentence_encoder)


def download_all():
    get_nltk_data()
    get_glove_emb(download_only=True)
    get_transformers("bert-base-cased")
    get_transformers("bert-base-uncased")
    get_transformers("gpt2-medium")
    get_universal_sentence_encoder()
    get_corenlp()
    get_stopwords()


if __name__ == "__main__":
    download_all()



import pandas as pd

from fibber.benchmark.benchmark_utils import load_detailed_result, update_overview_result
from fibber.benchmark.customized_metric_aggregation import customized_metric_for_nun_wins

# Columns where the number of wins should be computed.
# "L" means lower is better.
# "H" means higher is better.
COL_FOR_NUM_WINS = [
    ("USESemanticSimilarity_mean", "H"),
    ("GPT2GrammarQuality_mean", "L")
] + customized_metric_for_nun_wins

DATASET_NAME_COL = "0_dataset_name"
STRATEGY_NAME_COL = "1_paraphrase_strategy_name"


def make_overview():
    """Generate overview table from detailed table."""
    detailed_df = load_detailed_result()

    # verify detailed result
    for group_info, item in detailed_df.groupby([DATASET_NAME_COL, STRATEGY_NAME_COL]):
        assert len(item) == 1, (
            "Detailed results contains multiple runs for %s on %s." % (
                group_info[1], group_info[0]))

    results = {}
    for rid, item in detailed_df.iterrows():
        if item[STRATEGY_NAME_COL] not in results:
            model_name = item[STRATEGY_NAME_COL]
            tmp = dict()

            tmp[STRATEGY_NAME_COL] = item[STRATEGY_NAME_COL]
            for col_name, _ in COL_FOR_NUM_WINS:
                tmp[col_name] = 0

            results[model_name] = tmp

    for group_name, group in detailed_df.groupby(DATASET_NAME_COL):
        for _, r1 in group.iterrows():
            for _, r2 in group.iterrows():
                for column_name, direction in COL_FOR_NUM_WINS:
                    if ((direction == "H" and r1[column_name] > r2[column_name])
                            or (direction == "L" and r1[column_name] < r2[column_name])):
                        results[r1[STRATEGY_NAME_COL]][column_name] += 1

    update_overview_result(pd.DataFrame(list(results.values())))


if __name__ == "__main__":
    make_overview()



import datetime
import json

import numpy as np
import pandas as pd
import tqdm

from fibber import log
from fibber.metrics.bert_clf_prediction import BertClfPrediction
from fibber.metrics.editing_distance import EditingDistance
from fibber.metrics.glove_semantic_similarity import GloVeSemanticSimilarity
from fibber.metrics.gpt2_grammar_quality import GPT2GrammarQuality
from fibber.metrics.metric_base import MetricBase
from fibber.metrics.use_semantic_similarity import USESemanticSimilarity

logger = log.setup_custom_logger(__name__)


MEAN_AGG = "mean"
STD_AGG = "std"

SPECIAL_METRIC_AGGREGATION = {
    "BertClfPrediction": []
}
DEFAULT_AGGREGATION = [MEAN_AGG, STD_AGG]


def mean_aggregation_fn(x):
    """Compute mean of a list."""
    return float(np.mean(x))


def std_aggregation_fn(x):
    """Compute std of a list."""
    return float(np.std(x))


AGGREGATION_NAME_TO_FN = {
    MEAN_AGG: mean_aggregation_fn,
    STD_AGG: std_aggregation_fn
}


class MetricBundle(object):
    """MetricBundle can help easily initialize and compute multiple metrics."""

    def __init__(self,
                 use_editing_distance=True,
                 use_use_semantic_similarity=True,
                 use_glove_semantic_similarity=True,
                 use_gpt2_grammar_quality=True,
                 use_bert_clf_prediction=False,
                 customized_metrics=[],
                 **kargs):
        """Initialize various metrics.

        Args:
            use_editing_distance (bool): whether to use editing distance in the bundle.
            use_use_semantic_similarity (bool): whether to use Universal sentence encoder to
                compute sentence similarity
            use_glove_semantic_similarity (bool): whether to use Glove embeddings to compute
                sentence similarity.
            use_gpt2_grammar_quality (bool): whether to use GPT2 to compute sentence quality.
            use_bert_clf_prediction (bool): whether to include BERT classifier prediction in
                metrics.
            customized_metrics (list): a list of customized metrics.
            kargs: arguments for metrics. kargs will be passed to all metrics.
        """
        super(MetricBundle, self).__init__()

        assert isinstance(customized_metrics, list)
        for item in customized_metrics:
            assert isinstance(item, MetricBase)

        self._metrics = {}

        if use_editing_distance:
            metric = EditingDistance(**kargs)
            self._metrics[str(metric)] = metric
        if use_use_semantic_similarity:
            metric = USESemanticSimilarity(**kargs)
            self._metrics[str(metric)] = metric
        if use_glove_semantic_similarity:
            metric = GloVeSemanticSimilarity(**kargs)
            self._metrics[str(metric)] = metric
        if use_gpt2_grammar_quality:
            metric = GPT2GrammarQuality(**kargs)
            self._metrics[str(metric)] = metric
        if use_bert_clf_prediction:
            metric = BertClfPrediction(**kargs)
            self._metrics[str(metric)] = metric

        for metric in customized_metrics:
            assert str(metric) not in self._metrics, "Duplicate metric name."
            self._metrics[str(metric)] = metric

    def get_metric(self, metric_name):
        """Returns a metric in the bundle using the metric name.

        Metric name is the class name of a metric.

        Raises assertion error if metric is not found.

        Args:
            metric_name: the name of the matric.
        Returns:
            (object): a metric object.
        """
        assert metric_name in self._metrics
        return self._metrics[metric_name]

    def get_classifier_for_attack(self):
        """Returns the classifier for attack."""
        return self.get_metric("BertClfPrediction")

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Compute the results of all metrics in the bundle for one pair of text.

        Args:
            origin (str): original text.
            paraphrase (str): paraphrased text.
            data_record (str): the data record.
            paraphrase_field (str): choose from "text0", "text1".

        Returns:
            (dict):
        """
        ret = {}
        for name, metric in self._metrics.items():
            ret[name] = metric.measure_example(origin, paraphrase, data_record, paraphrase_field)
        return ret


def compute_metrics(metric_bundle, results, output_filename):
    """Compute the all metrics for results on a dataset.

    Args:
        metric_bundle (object): a MetricBundle object.
        results (dict): A fibber dataset with paraphrases.
        output_filename (str): A json filename to store results and metrics.

    Returns:
        (dict): the results dict with ``original_text_metrics`` and ``paraphrase_metrics`` added.
    """
    last_output_save_time = -1
    logger.info("Start measuring.")
    paraphrase_field = results["paraphrase_field"]

    for data_record in tqdm.tqdm(results["data"]):
        data_record_tmp = dict([(k, v) for k, v in data_record.items() if "_paraphrases" not in k])

        # Run metrics on original text
        data_record["original_text_metrics"] = metric_bundle.measure_example(
            data_record[paraphrase_field], data_record[paraphrase_field],
            data_record_tmp, paraphrase_field)

        # Run metrics on paraphrased text
        paraphrase_metric_list = []
        for paraphrase in data_record[paraphrase_field + "_paraphrases"]:
            paraphrase_metric_list.append(
                metric_bundle.measure_example(
                    data_record[paraphrase_field], paraphrase, data_record_tmp, paraphrase_field))

        data_record["paraphrase_metrics"] = paraphrase_metric_list

        # save tmp output every 30 seconds
        if datetime.datetime.now().timestamp() - last_output_save_time > 30:
            with open(output_filename, "w") as f:
                json.dump(results, f, indent=2)
            datetime.datetime.now().timestamp()

    with open(output_filename, "w") as f:
        json.dump(results, f, indent=2)

    return results


def aggregate_metrics(dataset_name, paraphrase_strategy_name, experiment_name, results,
                      customize_aggregation_fn_dict):
    """Aggregate paraphrase metrics on a dataset to a summary and store in a dict.

    Args:
        dataset_name (str): the name of the dataset.
        paraphrase_strategy_name (str): the name of the paraphrase strategy.
        experiment_name (str): the name of the experiment.
        results (dict): the fibber dataset with paraphrases and metrics. The return value of
            ``compute_metrics``.
        customize_aggregation_fn_dict (dict): A dict of customized aggregations. The dict is
            a mapping from aggregation name to aggregation function. The aggregation function
            should take one data_record, and returns a float.

    Returns:
        (dict): the aggregated metrics.
    """
    aggregated_result = pd.DataFrame()
    for data_record in results["data"]:
        aggregate_result_tmp = {}
        metric_df = pd.DataFrame(data_record["paraphrase_metrics"])
        for metric_name in metric_df.columns.tolist():
            metric_value = metric_df[metric_name].values
            if metric_name in SPECIAL_METRIC_AGGREGATION:
                aggregation_list = SPECIAL_METRIC_AGGREGATION[metric_name]
            else:
                aggregation_list = DEFAULT_AGGREGATION

            for aggregation in aggregation_list:
                aggregation_fn = AGGREGATION_NAME_TO_FN[aggregation]
                aggregated_value = aggregation_fn(metric_value)
                aggregate_result_tmp[metric_name + "_" + aggregation] = aggregated_value

        aggregate_result_tmp["ParaphrasesPerExample"] = len(data_record["paraphrase_metrics"])

        for name, fn in customize_aggregation_fn_dict.items():
            aggregate_result_tmp[name] = fn(data_record)

        aggregated_result = aggregated_result.append(aggregate_result_tmp, ignore_index=True)

    aggregated_result = dict(aggregated_result.mean(skipna=True))
    # hack column order by adding 0
    aggregated_result["0_dataset_name"] = dataset_name
    aggregated_result["1_paraphrase_strategy_name"] = paraphrase_strategy_name
    aggregated_result["2_experiment_name"] = experiment_name

    return aggregated_result



"""This module implements the paraphrase strategy using TextFooler."""


from fibber import log
from fibber.paraphrase_strategies.strategy_base import StrategyBase

logger = log.setup_custom_logger(__name__)

try:
    import textattack
    from textattack.attack_recipes.textfooler_jin_2019 import TextFoolerJin2019
    from textattack.models.wrappers.model_wrapper import ModelWrapper
except ImportError:
    logger.warning("TextAttack is not installed so TextFooler stategy can't be used. "
                   "Install it by `pip install textattack`.")
    ModelWrapper = object


class CLFModel(ModelWrapper):
    """A classifier wrapper for textattack package."""

    def __init__(self, clf_metric):
        self._clf_metric = clf_metric
        self._tokenizer = clf_metric._tokenizer
        self._context = None

    def __call__(self, text_list):
        ret = self._clf_metric.predict_dist_batch(text_list, self._context)
        return ret

    def set_context(self, context):
        self._context = context

    def tokenize(self, inputs):
        """Helper method that calls ``tokenizer.batch_encode`` if possible, and
        if not, falls back to calling ``tokenizer.encode`` for each input."""
        if hasattr(self._tokenizer, "batch_encode"):
            return self._tokenizer.batch_encode(inputs)
        else:
            return [self._tokenizer.encode(x) for x in inputs]


class TextFoolerStrategy(StrategyBase):
    """This strategy uses TextFooler to generate paraphrase.s

    We use TextFooler to attack the classifier in metric_bundle.
        If the attack succeeds, we use the adversarial sentence as a paraphrase.
        If the attack fails, we use the original sentence as a paraphrase.

    This strategy always returns one paraphrase for one data record, regardless of `n`.
    """

    __abbr__ = "tf"

    def fit(self, trainset):
        if ModelWrapper is None:
            raise RuntimeError("no internet connection.")

        self._model = CLFModel(self._metric_bundle.get_classifier_for_attack())
        self._textfooler = TextFoolerJin2019.build(self._model)

    def paraphrase_example(self, data_record, field_name, n):
        """Generate paraphrased sentences."""
        if field_name == "text1":
            self._model.set_context(data_record["text0"])

        att = next(self._textfooler.attack_dataset(
            [(data_record[field_name], data_record["label"])]))
        if isinstance(att, textattack.attack_results.SuccessfulAttackResult):
            return [att.perturbed_result.attacked_text.text]
        else:
            return [data_record[field_name]]



from fibber.metrics.editing_distance import EditingDistance
from fibber.metrics.glove_semantic_similarity import GloVeSemanticSimilarity
from fibber.metrics.gpt2_grammar_quality import GPT2GrammarQuality
from fibber.metrics.metric_utils import MetricBundle
from fibber.metrics.use_semantic_similarity import USESemanticSimilarity
from fibber.resources import get_nltk_data

get_nltk_data()


def test_editing_distance():
    editing_distance_metric = EditingDistance()

    s1 = "aa bb xx dd zz xx ee"
    s2 = "aa bb xy yz dd xy ee"
    assert editing_distance_metric.measure_example(s1, s2) == 4

    s1 = "aa bb cc dd"
    s2 = "aa bb dd"
    assert editing_distance_metric.measure_example(s1, s2) == 1

    s1 = "aa bb cc"
    s2 = "aa bb cc"
    assert editing_distance_metric.measure_example(s1, s2) == 0

    s1 = "a. b; c?"
    s2 = "a1, b& c'"
    assert editing_distance_metric.measure_example(s1, s2) == 1


def test_use_semantic_similarity():
    use_semantic_similarity_metric = USESemanticSimilarity()
    s1 = "Sunday is the first day in a week."
    s2 = "Obama was the president of the United State."
    metric1 = use_semantic_similarity_metric.measure_example(s1, s2)
    assert metric1 < 0.5

    s1 = "Sunday is the first day in a week"
    s2 = "Saturday is the last day in a week"
    metric2 = use_semantic_similarity_metric.measure_example(s1, s2)
    assert metric2 > 0.6

    s1 = "Sunday is the first day in a week"
    s2 = ["Obama was the president of the United State.",
          "Saturday is the last day in a week"]
    metric3 = use_semantic_similarity_metric.measure_batch(s1, s2)
    assert len(metric3) == 2
    assert abs(metric3[0] - metric1) < 1e-4
    assert abs(metric3[1] - metric2) < 1e-4


def test_gpt2_grammar_quality():
    gpt2_grammar_quality_metric = GPT2GrammarQuality()
    s1 = "Saturday is the last day in a week."
    s2 = "Sunday is is is first daay in a week."
    assert gpt2_grammar_quality_metric.measure_example(s1, s2) > 5

    s1 = "Sunday is the last day in a week."
    s2 = "Saturday is the last day in a week."
    assert gpt2_grammar_quality_metric.measure_example(s1, s2) < 2


def test_glove_semantic_similarity():
    glove_semantic_similarity_metric = GloVeSemanticSimilarity()

    s1 = "the a the to"
    s2 = "he him"
    assert glove_semantic_similarity_metric.measure_example(s1, s2) == 0

    s1 = "Saturday is the last day in a week"
    s2 = "Sunday is the last day in a week"
    assert 0.95 < glove_semantic_similarity_metric.measure_example(s1, s2) < 1


def test_metric_bundle():
    metric_bundle = MetricBundle()
    s1 = "Saturday is the last day in a week"
    s2 = "Sunday is the last day in a week"
    results = metric_bundle.measure_example(s1, s2)
    assert len(results) == 4
